# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals

import numpy as np
import declarative
import scipy.optimize

c_m_s = 299792458

def coupled_cavity_2ph_fitSQZdetune(F_Hz, **kwargs):
    def errfunc(det_B_rad):
        ret = coupled_cavity_2ph(F_Hz, det_B_rad = det_B_rad, **kwargs)
        select = F_Hz < 3e3
        SQZ_ang = ret.SQZ_ang[select]
        avg_SQZ_ang = np.mean(SQZ_ang)
        SQZ_ang_ERR = np.mean((SQZ_ang - avg_SQZ_ang)**2)
        return SQZ_ang_ERR
    out = scipy.optimize.minimize(fun = errfunc, x0 = (0.003,), method = 'Nelder-Mead')
    det_B_rad = out.x[0]

    ret = coupled_cavity_2ph(F_Hz, **kwargs)
    select = F_Hz < 3e3
    SQZ_ang = ret.SQZ_ang[select]
    avg_SQZ_ang = np.mean(SQZ_ang)
    SQZ_ang_ERR = np.mean(SQZ_ang - avg_SQZ_ang)**2
    return declarative.Bunch(locals())


def coupled_cavity_2ph(F_Hz, **kwargs):
    CC_p = coupled_cavity(F_Hz, **kwargs)
    CC_n = coupled_cavity(-F_Hz, **kwargs)
    A = np.array([[1, 1], [1j, -1j]]) / 2**.5
    A_inv = np.linalg.inv(A)

    REFL_AB_12 = A @ matrix_stack([
        [CC_p.REFL_AB[..., 0, 0], 0],
        [0, CC_n.REFL_AB[..., 0, 0].conjugate()],
    ]) @ A_inv

    TRANS_CA_12 = A @ matrix_stack([
        [CC_p.TRANS_CA[..., 0, 0], 0],
        [0, CC_n.TRANS_CA[..., 0, 0].conjugate()]
    ]) @ A_inv

    U, S, Vh = np.linalg.svd(REFL_AB_12)

    roto = CC_p.REFL_AB[..., 0, 0] * CC_n.REFL_AB[..., 0, 0]
    #RMS_imbalance = abs(abs(CC_p.REFL_AB[..., 0, 0])**2 - abs(CC_n.REFL_AB[..., 0, 0])**2)**0.5
    RMS_imbalance = abs(abs(CC_p.REFL_AB[..., 0, 0]) - abs(CC_n.REFL_AB[..., 0, 0])) / 2
    mean_loss = ((abs(CC_p.REFL_AB[..., 0, 0]) + abs(CC_n.REFL_AB[..., 0, 0])) / 2)**2
    #REFL_AB_12[..., 0, 1] / REFL_AB_12[..., 1, 1]
    SQZ_ang = np.angle(roto)

    REFL_LOSS = 1 - S[...,]**2
    return declarative.Bunch(locals())

def coupled_cavity(
    F_Hz       = 0,
    T_BC       = 0.014,
    T_AB       = 0.325,
    length_B_m = 56,
    length_C_m = 3995,
    det_B_rad  = 0,
    det_C_rad  = 0,
    L_C        = 35e-6,
    L_B        = .00,
    gouy_B     = -np.pi * .11,
    gouy_C     = +np.pi * .21,
    MM_L_BC    = 0.0,
    MM_phi_BC  = 0 * np.pi,
    MM_L_OMC   = 0.0,
    MM_L_AB    = 0.0,
    MM_phi_AB  = 0 * np.pi,
):
    dt_B = length_B_m / c_m_s
    dt_C = length_C_m / c_m_s

    phases_C = matrix_stack([
        [1, 0],
        [0, np.exp(1j * gouy_C)]
    ]) @ matrix_stack([
        [np.exp(1j * det_C_rad - 2j * np.pi * F_Hz * dt_C), 0],
        [0, np.exp(1j * det_C_rad - 2j * np.pi * F_Hz * dt_C)]
    ])

    phases_B = matrix_stack([
        [1, 0],
        [0, np.exp(1j * gouy_B)]
    ]) @ matrix_stack([
        [np.exp(1j * det_B_rad - 2j * np.pi * F_Hz * dt_B), 0],
        [0, np.exp(1j * det_B_rad - 2j * np.pi * F_Hz * dt_B)]
    ])

    MM_BC = matrix_stack([
        [(1 - MM_L_BC)**.5, np.exp(1j * MM_phi_BC) * MM_L_BC**.5],
        [-np.exp(-1j * MM_phi_BC) * MM_L_BC**.5, (1 - MM_L_BC)**.5]
    ])

    MM_AB = matrix_stack([
        [(1 - MM_L_AB)**.5, np.exp(1j * MM_phi_AB) * MM_L_AB**.5],
        [-np.exp(-1j * MM_phi_AB) * MM_L_AB**.5, (1 - MM_L_AB)**.5]
    ])

    MM_OMC = matrix_stack([
        [(1 - MM_L_OMC)**.5, MM_L_OMC**.5],
        [-MM_L_OMC**.5, (1 - MM_L_OMC)**.5]
    ])

    id2 = np.eye(2)

    r_CD = (1 - L_C)**.5

    t_BC = T_BC**.5
    r_BC = (1 - T_BC)**.5

    t_AB = T_AB**.5
    r_AB = (1 - T_AB)**.5

    RT_C = phases_C @ phases_C * r_CD * r_BC

    REFL_BC = (r_BC*id2 - MM_BC @ (t_BC**2/r_BC * RT_C) @ np.linalg.inv(id2 - RT_C) @ np.linalg.inv(MM_BC))
    TRANS_CB = MM_BC @ (t_BC * phases_C * id2) @ np.linalg.inv(id2 - RT_C) @ phases_C

    RT_B = r_AB * (phases_B @ REFL_BC @ phases_B) * (1 - L_B)**0.5
    REFL_AB = r_AB * id2 - MM_AB @ ((t_AB**2 * (RT_B/r_AB)) @ np.linalg.inv(id2 - RT_B) @ np.linalg.inv(MM_AB) @ np.linalg.inv(MM_OMC))

    TRANS_CA = MM_AB @ (t_AB * phases_B) @ np.linalg.inv(id2 - RT_B) @ TRANS_CB

    return declarative.Bunch(locals())


def matrix_stack(arr, **kwargs):
    Nrows = len(arr)
    Ncols = len(arr[0])
    vals = []
    dtypes = []
    for r_idx, row in enumerate(arr):
        assert(len(row) == Ncols)
        for c_idx, kdm in enumerate(row):
            kdm = np.asarray(kdm)
            vals.append(kdm)
            dtypes.append(kdm.dtype)

    #dt = np.find_common_type(dtypes, ())
    dt = np.result_type(*vals)
    bc = np.broadcast(*vals)

    if len(bc.shape) == 0:
        return np.array(arr)

    Marr = np.empty(bc.shape + (Nrows, Ncols), dtype = dt, **kwargs)
    #print(Marr.shape)

    for r_idx, row in enumerate(arr):
        for c_idx, kdm in enumerate(row):
            Marr[..., r_idx, c_idx] = kdm
    return Marr
